package at.cosimo.verocl.util.ecore;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;


public class GeneralTraverser {
	
	private Map<EClass, Map<String, EStructuralFeature>> estructuralMap = 
			new HashMap<>();
		
			
	public EStructuralFeature getFeature(EClass clazz, String name) {
		Map<String, EStructuralFeature> map = estructuralMap.get(clazz);
		if (map == null) {
			estructuralMap.put(clazz, map = new HashMap<>());
		}
		EStructuralFeature ret = map.get(name);
		if (ret == null && clazz != null) {
			map.put(name, ret = clazz.getEStructuralFeature(name));
		}
		return ret;
	}
	
	public interface Filter extends EobjFilter {
		public boolean doesApply(Object obj);
	}
	
	public interface EobjFilter {
		public boolean doesApply(EObject obj);
	}
	
	public interface Transformer {
		public Object transform(Object obj);
	}
	
	public interface EobjTransformer {
		public Object transform(EObject obj);
	}
	
	
	public abstract class TraverserIFace {
		
		public Object getPValue(String... path) {
			TraverserIFace cur = this;
			for (String str: path) {
				cur = cur.attr(str);
			}
			return cur.getValue();
		}
		
		public abstract TraverserIFace filter(Filter filter);
		
		public abstract TraverserIFace filter(EobjFilter filter);
		
		public abstract TraverserIFace transform(EobjTransformer filter);
		
		public abstract TraverserIFace transform(Transformer filter);
		
		public abstract TraverserIFace attr(String name);
		
		public abstract Object getValue();
	}
	
	public class NullTraverser extends TraverserIFace {
		
		public String toString() {
			return "null";
		}

		@Override
		public NullTraverser attr(String name) {
			return this;
		}
		
		public Object getValue() {
			return null;
		}
		
		public NullTraverser filter(Filter filter) {
			return this;
		}
		
		public NullTraverser filter(EobjFilter filter) {
			return this;
		}

		@Override
		public TraverserIFace transform(EobjTransformer filter) {
			return this;
		}

		@Override
		public TraverserIFace transform(Transformer filter) {
			return this;
		}
		
	}
	
	public class MultiTraverser extends TraverserIFace {
		private List<EObject> thiz = new ArrayList<>();
		
		public MultiTraverser(List<EObject> obj) {
			this.thiz = obj;
		}

		public MultiTraverser filter(Filter filter) {
			return filter((EobjFilter)filter);
		}
		
		public MultiTraverser filter(EobjFilter filter) {
			
			List<EObject> ret = new ArrayList<>();
			Iterator<EObject> iter = thiz.iterator();
			while (iter.hasNext()) {
				EObject obj = iter.next();
				if (filter.doesApply(obj)) {
					ret.add(obj);
				}
			}
			return new MultiTraverser(ret);
		}
		
		@Override
		public TraverserIFace transform(EobjTransformer filter) {
			List<Object> ret = new ArrayList<>();
			Iterator<EObject> iter = thiz.iterator();
			boolean onlyEObj = true;
			while (iter.hasNext()) {
				EObject obj = iter.next();
				Object toAdd = filter.transform(obj);
				if (!(toAdd instanceof EObject)) {
					onlyEObj = false;
				}
				ret.add(toAdd);
			}
			return onlyEObj?new MultiTraverser((List)ret):new MultiValueTraverser(ret);
		}

		@Override
		public TraverserIFace transform(Transformer filter) {
			return transform(filter);
		}
		
		@Override
		public TraverserIFace attr(String name) {
			List<Object> ret = new ArrayList<>();
			boolean onlyEObj = true;
			for (EObject obj: thiz) {
				for (Object sub: getObjects(obj, name)) {
					if (!(sub instanceof EObject)) {
						onlyEObj = false;
					}
					ret.add(sub);
				}
			}
			if (onlyEObj) {
				MultiTraverser rett = new MultiTraverser((List)ret);
				return rett;
			} else {
				MultiValueTraverser rett = new MultiValueTraverser((List)ret);
				return rett;
			}
		}

		@Override
		public List<EObject> getValue() {
			return thiz;
		}

		
		
	}
	
	public class MultiValueTraverser extends TraverserIFace {
		private List<Object> thiz = new ArrayList<>();

		public String toString() {
			StringBuilder ret = new StringBuilder();
			boolean first = true;
			ret.append("MultiValue [");
			for (Object o: thiz) {
				if (first) {first = false;} else {ret.append(", ");}
				ret.append(String.valueOf(o));
			}
			ret.append("]");
			return ret.toString();
		}
		
		public MultiValueTraverser(List<Object> ret) {
			thiz = ret;
		}

		@Override
		public NullTraverser attr(String name) {
			return new NullTraverser();
		}

		@Override
		public List<Object> getValue() {
			return thiz;
		}

		@Override
		public TraverserIFace filter(Filter filter) {
			List<Object> ret = new ArrayList<>();
			Iterator<Object> iter = thiz.iterator();
			while (iter.hasNext()) {
				Object obj = iter.next();
				if (filter.doesApply(obj)) {
					ret.add(obj);
				}
			}
			return new MultiValueTraverser(ret);
		}

		@Override
		public TraverserIFace filter(EobjFilter filter) {
			List<Object> ret = new ArrayList<>();
			Iterator<Object> iter = thiz.iterator();
			while (iter.hasNext()) {
				Object objj = iter.next();
				if (!(objj instanceof EObject)) {continue;}
				EObject obj = (EObject)objj;
				if (filter.doesApply(obj)) {
					ret.add(obj);
				}
			}
			return new MultiValueTraverser(ret);			
		}

		@Override
		public TraverserIFace transform(EobjTransformer filter) {
			List<Object> ret = new ArrayList<>();
			Iterator<Object> iter = thiz.iterator();
			boolean onlyEObj = true;
			while (iter.hasNext()) {
				Object objj = iter.next();
				if (!(objj instanceof EObject)) {continue;}
				EObject obj = (EObject)objj;
				Object toAdd = filter.transform(obj);
				if (!(toAdd instanceof EObject)) {
					onlyEObj = false;
				}
				ret.add(toAdd);
			}
			return onlyEObj?new MultiTraverser((List)ret):new MultiValueTraverser(ret);	
		}

		@Override
		public TraverserIFace transform(Transformer filter) {
			List<Object> ret = new ArrayList<>();
			Iterator<Object> iter = thiz.iterator();
			boolean onlyEObj = true;
			while (iter.hasNext()) {
				Object obj = iter.next();
				Object toAdd = filter.transform(obj);
				if (!(toAdd instanceof EObject)) {
					onlyEObj = false;
				}
				ret.add(toAdd);
			}
			return onlyEObj?new MultiTraverser((List)ret):new MultiValueTraverser(ret);	
		}
		
	}
	
	public class ValueTraverser extends TraverserIFace {
		Object thiz;
		
		public String toString() {
			return String.valueOf(thiz);
		}
		
		public ValueTraverser(Object thiz) {
			this.thiz = thiz;
		}
		
		@Override
		public NullTraverser attr(String name) {
			return new NullTraverser();
		}

		@Override
		public Object getValue() {
			return thiz;
		}

		@Override
		public TraverserIFace filter(Filter filter) {
			if (filter.doesApply(thiz)) {
				return this;
			}
			return new NullTraverser();
		}

		@Override
		public TraverserIFace filter(EobjFilter filter) {
			if ((thiz instanceof EObject) && filter.doesApply((EObject)thiz)) {
				return this;
			}
			return new NullTraverser();
		}

		@Override
		public TraverserIFace transform(EobjTransformer filter) {
			if ((thiz instanceof EObject)) {
				Object ret = filter.transform((EObject)thiz);
				if (ret instanceof EObject) {
					return new SimpleTraverser((EObject)ret);
				} else {
					return new ValueTraverser(ret);
				}
			}
			return new NullTraverser();
		}

		@Override
		public TraverserIFace transform(Transformer filter) {
			Object ret = filter.transform(thiz);
			if (ret instanceof EObject) {
				return new SimpleTraverser((EObject)ret);
			} else {
				return new ValueTraverser(ret);
			}
		}
		
	}
	
	public class SimpleTraverser extends TraverserIFace {
		private EObject thiz;
		
		public SimpleTraverser(EObject thiz) {
			this.thiz = thiz;
		}
		
		public String toString() {
			return "Traverser for obj of class " + ((thiz==null)?"null":((thiz.eClass()!=null)?thiz.eClass().getName():"null"));
		}

		@Override
		public TraverserIFace attr(String name) {
			if (thiz == null) {
				return new NullTraverser();
			}
			List<Object> ret = new ArrayList<>();
			EStructuralFeature feat = getFeature(thiz.eClass(), name);
			if (feat != null) {
				if (feat.getUpperBound() >= 0 && feat.getUpperBound() <= 1) {
					Object fval = thiz.eGet(feat);
					if (fval instanceof EObject) {
						return new SimpleTraverser((EObject)fval);
					} else {
						return new ValueTraverser(fval);
					}
				} else {
					Collection fval = (Collection)thiz.eGet(feat);
					boolean onlyEObj = true;
					for (Object sub: fval) {
						if (!(sub instanceof EObject)) {
							onlyEObj = false;
						}
						ret.add(sub);
					}
					if (onlyEObj) {
						MultiTraverser rett = new MultiTraverser((List)ret);
						return rett;
					} else {
						MultiValueTraverser rett = new MultiValueTraverser((List)ret);
						return rett;
					}
				}
			}
			return new NullTraverser();
		}

		@Override
		public EObject getValue() {
			return thiz;
		}
		
		@Override
		public TraverserIFace filter(Filter filter) {
			return filter((EobjFilter)filter);
		}

		@Override
		public TraverserIFace filter(EobjFilter filter) {
			if (filter.doesApply(thiz)) {
				return this;
			}
			return new NullTraverser();
		}

		@Override
		public TraverserIFace transform(EobjTransformer filter) {
			if ((thiz instanceof EObject)) {
				Object ret = filter.transform((EObject)thiz);
				if (ret instanceof EObject) {
					return new SimpleTraverser((EObject)ret);
				} else {
					return new ValueTraverser(ret);
				}
			}
			return new NullTraverser();
		}

		@Override
		public TraverserIFace transform(Transformer filter) {
			Object ret = filter.transform(thiz);
			if (ret instanceof EObject) {
				return new SimpleTraverser((EObject)ret);
			} else {
				return new ValueTraverser(ret);
			}
		}
				
	}
	
	public SimpleTraverser traverser(EObject obj) {
		return new SimpleTraverser(obj);
	}
	
	public MultiTraverser traverser(EObject... obj) {
		List<EObject> thiz = new ArrayList<>();
		for (EObject o: obj) {
			thiz.add(o);
		}
		return new MultiTraverser(thiz);
	}
	
	/**Does not copy the list*/
	public MultiTraverser traverser(List<EObject> obj) {
		MultiTraverser ret = new MultiTraverser(obj);
		return ret;
	}
			
	public Object getObject(EObject from, String attr) {
		if (from == null) {
			return null;
		}
		EStructuralFeature feat = getFeature(from.eClass(),attr);
		if (feat == null) {
			System.err.println("Could not retrieve feature " + attr + " for class " + from.eClass());
			return null;
		}
		return from.eGet(feat);
	}
	
	public Collection<Object> getObjects(EObject from, String attr) {
		if (from == null) {
			return null;
		}
		EStructuralFeature feat = getFeature(from.eClass(),attr);
		if (feat == null) {
			System.err.println("Could not retrieve feature " + attr + " for class " + from.eClass());
			return null;
		}
		if (feat.getUpperBound() < 0 || feat.getUpperBound() > 1) {
			return (Collection)from.eGet(feat);
		} else {
			return Collections.singleton(from.eGet(feat));
		}
	}

}
